import json
import xml.etree.ElementTree as ET
import xmltodict


def is_string_valid_xml(xml_string):
    """validates a string is valid xml

    Args:
        xml_string (str): string validating xml

    Returns:
        bool: true if the string is valid xml otherwise false
    """
    try:
        ET.fromstring(xml_string)
        return True
    except ET.ParseError:
        return False


def convert_json_to_dict(json_string: str) -> dict:
    """
    Convert JSON string to dictionary.

    Args: json_string (str): The JSON string.

    Returns: dict: The converted dictionary.
    """
    return json.loads(json_string)


def convert_xml_to_dict(xml_string: str) -> dict:
    """
    Convert XML string to dictionary.

    Args:  xml_string (str): The XML string.

    Returns: dict: The converted dictionary.
    """
    return xmltodict.parse(xml_string)


def convert_dict_to_json(input_dict: dict) -> str:
    """
    Convert dictionary to JSON string.

    Args: input_dict (dict): The dictionary.

    Returns: str: The converted JSON string.
    """
    return json.dumps(input_dict)



from typing import Dict

PARSER_MAPPING: Dict[str, Dict[str, str]] = {
    "default_parser": {
        "module": "ncpeek.parsers.default_parser",
        "class": "DefaultParser",
    },
    "cisco_xe_ietf-interfaces.xml": {
        "module": "ncpeek.parsers.cisco_xe_ietf_interfaces",
        "class": "InterfaceStatsIETF_IOSXEParser",
    },
    "Cisco-IOS-XE-interfaces-oper.xml": {
        "module": "ncpeek.parsers.cisco_ios_xe_interfaces_oper",
        "class": "InterfaceStatsIOSXEParser",
    },
    "Cisco-IOS-XE-memory-oper.xml": {
        "module": "ncpeek.parsers.cisco_ios_xe_memory_oper",
        "class": "CiscoIOSXEMemoryParser",
    },
    "http://cisco.com/ns/yang/Cisco-IOS-XE-isis-oper:/isis-oper-data/isis-instance": {
        "module": "ncpeek.parsers.cisco_ios_xe_isis_oper",
        "class": "ISISStatsIOSXEParser",
    },
}



from typing import Optional
from dataclasses import dataclass, field
from ncpeek.netconf_parsers import Parser
from ncpeek.netconf_devices import NetconfDevice


@dataclass
class ISISStatsIOSXEParser(Parser):
    """
    A parser for ISIS stats for IOSXE devices.
    """

    device: NetconfDevice = None
    isis_interfaces_adj_up: int = 0
    stats: list = field(default_factory=list)
    netconf_filter_id: Optional[str] = None

    def parse(
        self,
        data_to_parse: dict,
        device: NetconfDevice,
        netconf_filter_id: str,
    ) -> list[dict]:
        """
        Parse ISIS stats data.

        Args:
            data (Dict): The data to parse.
            device (NetconfDevice): The device the data is related to.
            netconf_filter_id (str): The filter ID used for netconf.

        Returns:
            List[Dict]: The parsed data.
        """
        self.device = device
        self._extract_neighbor_stats(data=data_to_parse)
        self.netconf_filter_id = netconf_filter_id
        return self.stats

    def _extract_neighbor_stats(self, data: dict) -> None:
        isis_instances: dict = data["data"]["isis-oper-data"]["isis-instance"]

        for key, isis_instance in isis_instances.items():
            if "isis-neighbor" in key:
                self._count_adjcencies(instance=isis_instance)
                self._extract_metadata(instance=isis_instance)

    def _count_adjcencies(self, instance: list) -> None:
        """If you have 2 isis interfaces data comes inside a list.
        if you shutdown one interface and only 1 is up,
        the data will not come insdie a list."""
        if isinstance(instance, list):
            for neighbor in instance:
                if "isis-adj-up" in neighbor["state"]:
                    self.isis_interfaces_adj_up += 1
        else:
            if "isis-adj-up" in instance["state"]:
                self.isis_interfaces_adj_up += 1

    def _extract_metadata(self, instance: list) -> None:
        if isinstance(instance, list):
            for neighbor in instance:
                self.stats.append(
                    self._get_neighbor_metadata(neighbor=neighbor)
                )
        else:
            self.stats.append(self._get_neighbor_metadata(neighbor=instance))

    def _get_neighbor_metadata(self, neighbor: dict) -> dict:
        return {
            "isis_interfaces_adj_up": self.isis_interfaces_adj_up,
            "isis_status": neighbor["state"],
            "neighbor_id": neighbor["system-id"].replace(" ", "_"),
            "field": self.netconf_filter_id,
            "device": self.device.hostname,
            "ip": self.device.host,
            "interface_name": neighbor["if-name"],
            "ipv4_address": neighbor["ipv4-address"],
        }



from ncpeek.netconf_parsers import Parser
from ncpeek.factory.factory_mappings import PARSER_MAPPING


def get_parser(netconf_filter: str) -> Parser:
    """
    Returns a parser instance based on the provided Netconf filter.

    Args: netconf_filter: The Netconf filter ID.
    Returns: An instance of the appropriate parser.
    """

    parser_config = PARSER_MAPPING.get(
        netconf_filter, PARSER_MAPPING["default_parser"]
    )

    parser_module = __import__(
        parser_config["module"], fromlist=[parser_config["class"]]
    )
    # The last () in the line parser_instance is used to instantiate the class.
    parser_instance = getattr(parser_module, parser_config["class"])()

    return parser_instance



import os
from pathlib import Path

DEFAULT_DEVICES_DIR = "devices"
DEFAULT_FILTER_DIR = "filters"


def read_settings(filename: str) -> str:
    """Read device settings"""
    file = resolve_path(filename=filename, kind="settings")
    return read_file(filename=file)


def read_filter(filename: str) -> str:
    """Read filter xml file"""
    file = resolve_path(filename=filename, kind="filter")
    return read_file(filename=file)


def resolve_path(filename: str, kind: str) -> str:
    """Resolve the path for files."""
    path = resolve_absolute_path(filename=filename)
    if is_valid_file_path(filepath=path):
        return path
    if "settings" in kind:
        return construct_default_devices_path(filename=filename)
    return construct_default_filter_path(filename=filename)


def resolve_absolute_path(filename: str) -> str:
    """Considers symlinks and home directory relative paths"""
    return os.path.realpath(os.path.expanduser(filename))


def is_valid_file_path(filepath: str) -> bool:
    """Check if a filepath is an existing file."""
    return os.path.exists(filepath) and os.path.isfile(filepath)


def construct_default_devices_path(filename: str) -> str:
    """Construct the default path for a devices file."""
    return f"{get_script_directory()}/{DEFAULT_DEVICES_DIR}/{filename}"


def construct_default_filter_path(filename: str) -> str:
    """Construct the default path for a filter file."""
    return f"{get_script_directory()}/{DEFAULT_FILTER_DIR}/{filename}"


def get_script_directory() -> str:
    """Get the directory of the script."""
    return Path(os.path.abspath(os.path.dirname(__file__))).parents[0]


def read_file(filename: str) -> str:
    """Read content from a file."""
    try:
        with open(filename, "r", encoding="utf-8") as file:
            return file.read()
    except Exception as err:
        raise ValueError(f"Unable to open file. Error: {err=}") from err


def append_to_file(filename: str, text: str) -> None:
    """Append text to a file."""
    try:
        with open(filename, "a", encoding="utf-8") as file:
            file.write(text)
    except Exception as err:
        raise ValueError(f"Unable to write to file. Error: {err=}") from err


def remove_path_from_filename(filename: str) -> str:
    """Remove the path from a filename, keeping the base name.
    If filename starts with http:// or https://, return as is."""
    if "http://" in filename or "https://" in filename:
        return filename
    return os.path.basename(filename)



from ncpeek.netconf_parsers import Parser
from ncpeek.netconf_devices import NetconfDevice
from ncpeek.parsers.remove_namespaces import (
    remove_namespaces_from_dict,
)


class DefaultParser(Parser):
    """
    A default parser which translate the rcp reply to a generic structure.
    """

    def parse(
        self,
        data_to_parse: dict,
        device: NetconfDevice,
        netconf_filter_id: str,
    ) -> list[dict]:
        """
        Parse data to a generic structure without any logic.

        Args:
            data_to_parse (Dict): The data to parse.
            device (NetconfDevice): The device the data is related to.
            netconf_filter_id (str): The filter ID used for netconf.

        Returns:
            List[Dict]: The parsed data.
        """
        data = remove_namespaces_from_dict(data=data_to_parse["data"])
        return [
            {
                "ip": device.host,
                "device": device.host,
                "field": netconf_filter_id,
                "data": data,
            }
        ]



from typing import Optional
from collections import ChainMap
from dataclasses import dataclass
from ncpeek.netconf_parsers import Parser
from ncpeek.netconf_devices import NetconfDevice


@dataclass
class InterfaceStatsIOSXEParser(Parser):
    """A parser for interface stats for IOSXE devices."""

    device: NetconfDevice = None
    netconf_filter_id: Optional[str] = None

    def parse(
        self,
        data_to_parse: dict,
        device: NetconfDevice,
        netconf_filter_id: str,
    ) -> list[dict]:
        """
        Parse interface stats data.

        Args:
            data (Dict): The data to parse.
            device (NetconfDevice): The device the data is related to.
            netconf_filter_id (str): The filter ID used for netconf.

        Returns:
            List[Dict]: The parsed data.
        """
        self.device = device
        self.netconf_filter_id = netconf_filter_id
        return self._interface_stats(data=data_to_parse)

    def _interface_stats(self, data: dict) -> list[dict]:
        stats: list = []
        interfaces: dict = data["data"]["interfaces"]["interface"]

        for interface in interfaces:
            metadata = self._extract_metadata(interface=interface)
            intf_stats = self._extract_statistics(
                interface=interface["statistics"]
            )
            stats.append(dict(ChainMap(metadata, intf_stats)))
        return stats

    def _extract_metadata(self, interface: dict) -> dict:
        return {
            "operational_status": 1
            if interface["oper-status"] == "if-oper-state-ready"
            else 0,
            "name": interface["name"].replace(" ", "_"),
            "field": self.netconf_filter_id,
            "device": self.device.hostname,
            "ip": self.device.host,
        }

    def _extract_statistics(self, interface: dict) -> dict:
        return {
            "in_octets": int(interface["in-octets"]),
            "in_errors": int(interface["in-errors"]),
            "out_octets": int(interface["out-octets"]),
            "out_errors": int(interface["out-errors"]),
            "in-broadcast-pkts": int(interface["in-broadcast-pkts"]),
            "in-crc-errors": int(interface["in-crc-errors"]),
            "in-discards": int(interface["in-discards"]),
            "in-discards-64": int(interface["in-discards-64"]),
            "in-errors-64": int(interface["in-errors-64"]),
            "in-multicast-pkts": int(interface["in-multicast-pkts"]),
            "in-unicast-pkts": int(interface["in-unicast-pkts"]),
            "in-unknown-protos": int(interface["in-unknown-protos"]),
            "in-unknown-protos-64": int(interface["in-unknown-protos-64"]),
            "num-flaps": int(interface["num-flaps"]),
            "out-broadcast-pkts": int(interface["out-broadcast-pkts"]),
            "out-discards": int(interface["out-discards"]),
            "out-multicast-pkts": int(interface["out-multicast-pkts"]),
            "out-octets-64": int(interface["out-octets-64"]),
            "out-unicast-pkts": int(interface["out-unicast-pkts"]),
            "rx-kbps": int(interface["rx-kbps"]),
            "rx-pps": int(interface["rx-pps"]),
            "tx-kbps": int(interface["tx-kbps"]),
            "tx-pps": int(interface["tx-pps"]),
        }



from typing import Optional
from dataclasses import dataclass
from ncpeek.utils.file_utils import (
    read_filter,
    read_settings,
    remove_path_from_filename,
)
from ncpeek.utils.text_utils import (
    is_string_valid_xml,
    convert_json_to_dict,
)
from ncpeek.args.xpath_parse import extract_xpath
from ncpeek.args.arg_parser import create_argument_parser


@dataclass
class SettingsParser:
    """
    A class to parse and handle settings for netconf client.
    """

    _filter_id: Optional[str] = None
    _network_filter: Optional[str] = None
    _device_settings: Optional[str] = None

    def parse_arguments(self) -> None:
        """Parse command-line arguments and set device settings and filters."""
        args = create_argument_parser().parse_args()
        self._device_settings = self._load_settings(args.device_settings)

        if args.xml_filter:
            self._parse_xml_filter(args.xml_filter)
        elif args.xpath_filter:
            self._parse_xpath_filter(args.xpath_filter)

    def set_device_settings(self, device_settings: str) -> None:
        """Set device settings from a json file."""
        self._device_settings = self._load_settings(device_settings)

    def set_xml_filter(self, xml_filter: str) -> None:
        """Set XML filter from a provided string."""
        self._parse_xml_filter(xml_filter)

    def set_xpath_filter(self, xpath_filter: str) -> None:
        """Set XPath filter from a provided string."""
        self._parse_xpath_filter(xpath_filter)

    def get_device_settings(self) -> str:
        """Get the device settings."""
        if not self._device_settings:
            raise ValueError("Device Settings not provided")
        return self._device_settings

    def get_filter_id(self) -> str:
        """Get the filter id."""
        if not self._filter_id:
            raise ValueError("filter not provided")
        return self._filter_id

    def get_netconf_filter(self) -> str:
        """Get the network filter."""
        return self._network_filter

    def _parse_xml_filter(self, filter_id: str) -> None:
        """Parse XML filter from a provided string."""
        if is_string_valid_xml(xml_string=filter_id):
            self._filter_id = "generic"
            self._network_filter = filter_id
            return

        try:
            self._set_filter_id(filter_id=filter_id)
            network_filter = read_filter(filename=filter_id)
        except Exception as err:
            raise ValueError(
                f"Error opening file {err=}. Make sure a valid filename is provided for a xml filter."
            ) from err
        if is_string_valid_xml(xml_string=network_filter):
            self._network_filter = network_filter
        else:
            raise ValueError("No valid XML found in file/string")

    def _parse_xpath_filter(self, filter_id: str) -> None:
        """Parse XPath filter from a provided string."""
        self._set_filter_id(filter_id=filter_id)
        self._network_filter = extract_xpath(netconf_filter=filter_id)

    def _load_settings(self, device_settings: str) -> dict:
        """Load settings from a provided string."""
        if isinstance(device_settings, list):
            return device_settings
        if isinstance(device_settings, str):
            try:
                # passing json directly
                return convert_json_to_dict(json_string=device_settings)
            except Exception:
                return convert_json_to_dict(
                    read_settings(filename=device_settings)
                )

    def _set_filter_id(self, filter_id: str) -> str:
        self._filter_id = remove_path_from_filename(filename=filter_id)



"""
Module for removing namespaces from rpc-reply.
namespaces are keys on a dictionary
"""

from collections import deque
from typing import Any, Dict

KEYS_TO_REMOVE = {"@xmlns", "@xmlns:nc"}


def remove_namespaces_from_dict(
    data: Dict[str, Any], depth_limit=10
) -> Dict[str, Any]:
    """
    Remove specified keys from a dictionary up to a certain depth.

    Args:
        data (Dict[str, Any]): The original dictionary.
        depth_limit (int): The maximum depth to search into the dictionary.
                           Default is 10.
                           After the depth is reached, the remaining
                           dictionaries are copied and not processed.

    Returns:
        Dict[str, Any]: A new dictionary without the specified keys.
    """
    queue = deque([((), data)])
    result = {}

    while queue:
        path, current_data = queue.popleft()
        depth = len(path)

        for key, value in current_data.items():
            if key in KEYS_TO_REMOVE and depth < depth_limit:
                continue

            new_path = path + (key,)

            if isinstance(value, dict) and depth < depth_limit - 1:
                queue.append((new_path, value))
            else:
                current_level = result
                for sub_key in new_path[:-1]:
                    current_level = current_level.setdefault(sub_key, {})
                current_level[new_path[-1]] = value

    return result



from typing import Optional
from dataclasses import dataclass
from ncpeek.netconf_parsers import Parser
from ncpeek.netconf_devices import NetconfDevice


@dataclass
class InterfaceStatsIETF_IOSXEParser(Parser):
    """
    A parser for interface stats for IETF IOSXE devices.
    """

    device: NetconfDevice = None
    netconf_filter_id: Optional[str] = None

    def parse(
        self,
        data_to_parse: dict,
        device: NetconfDevice,
        netconf_filter_id: str,
    ) -> list[dict]:
        """
        Parse interface stats data.

        Args:
            data (Dict): The data to parse.
            device (NetconfDevice): The device the data is related to.
            netconf_filter_id (str): The filter ID used for netconf.

        Returns:
            List[Dict]: The parsed data.
        """
        self.device = device
        self.netconf_filter_id = netconf_filter_id
        return self._extract_interface_stats(data=data_to_parse)

    def _extract_interface_stats(self, data: dict) -> list[dict]:
        stats: list = []
        interfaces: dict = data["data"]["interfaces-state"]["interface"]

        for interface in interfaces:
            stats.append(self._get_interface_stats(interface=interface))
        return stats

    def _get_interface_stats(self, interface: dict) -> dict:
        return {
            "operational_status": 1 if interface["oper-status"] == "up" else 0,
            "in_octets": int(interface["statistics"]["in-octets"]),
            "in_errors": int(interface["statistics"]["in-errors"]),
            "out_octets": int(interface["statistics"]["out-octets"]),
            "out_errors": int(interface["statistics"]["out-errors"]),
            "name": interface["name"].replace(" ", "_"),
            "field": self.netconf_filter_id,
            "device": self.device.hostname,
            "ip": self.device.host,
        }



from typing import Optional, Union
from dataclasses import dataclass
from ncpeek.utils.text_utils import (
    convert_xml_to_dict,
    convert_dict_to_json,
)
from ncpeek.netconf_devices import NetconfDevice
from ncpeek.args.parse_settings import SettingsParser
from ncpeek.factory.factory_parsers import get_parser
from ncpeek.netconf_session import NetconfSession

DEFAULT_NETCONF_OPERATION = "fetch"


@dataclass
class NetconfClient:
    """
    Main class for handling NetconfClient operations.
    """

    _settings = SettingsParser()
    _operation: Optional[str] = DEFAULT_NETCONF_OPERATION
    _filter_id: Optional[str] = None
    _netconf_filter: Optional[str] = None

    def execute_cli(self) -> str:
        """Executes command-line interface."""
        self._settings.parse_arguments()
        return self._run()

    def set_devices_settings(self, device_settings: Union[list, str]) -> None:
        """
        API: sets devices settings directly.
        Can be a python list, a json string or a string
        with the name of the json file.
        See examples under ncpeek/devices
        """
        self._settings.set_device_settings(device_settings)

    def set_xml_filter(self, xml_filter: str) -> None:
        """
        API: sets XML filter directly.
        Can be the filename of a xml file (relative and absolute paths accepted)
        or can be a python string with valid xml.
        """
        self._settings.set_xml_filter(xml_filter)

    def set_xpath_filter(self, xpath_filter: str) -> None:
        """API: sets XPath filter directly."""
        self._settings.set_xpath_filter(xpath_filter)

    def fetch(self) -> str:
        """Fetchs data from network device

        Returns:
            str: rpc reply from network device
        """
        self._operation = "fetch"
        return self._run()

    def _run(self) -> str:
        """Runs the main operations and returns results in JSON format."""

        devices = self._settings.get_device_settings()
        self._filter_id = self._settings.get_filter_id()
        self._netconf_filter = self._settings.get_netconf_filter()

        results = []
        for device in devices:
            results += self._process_device(device)

        return convert_dict_to_json(results)

    def _process_device(self, device: dict) -> dict:
        """Processes a single device operation and parsing its reply."""
        try:
            device = NetconfDevice(**device)
            rpc = NetconfSession(
                device=device,
                netconf_filter=self._netconf_filter,
                operation=self._operation,
            )
            data_dict = convert_xml_to_dict(xml_string=rpc.reply())
            parser = get_parser(netconf_filter=self._filter_id)
            parsed_data = parser.parse(
                data_to_parse=data_dict,
                device=device,
                netconf_filter_id=self._filter_id,
            )
        except Exception as err:
            parsed_data = [{"error": f"{err=}"}]
        return parsed_data


def cli() -> None:
    """Runs the netconf client using the CLI with the arguments supplied."""
    client = NetconfClient()
    try:
        print(client.execute_cli())
    except Exception as err:
        print(f"Error found: {err=}")


if __name__ == "__main__":
    cli()



from typing import Optional
from dataclasses import dataclass
from ncclient import manager
from ncpeek.netconf_devices import NetconfDevice


@dataclass
class NetconfSession:
    """Handles the netconf session for a device."""

    device: NetconfDevice
    netconf_filter: str
    operation: str
    result: Optional[str] = None

    def __post_init__(self) -> None:
        """
        Executes the netconf operation.

        Returns: The result of the Netconf  operation.
        """
        self.result = self._perform_operation()

    def reply(self) -> str:
        """retrieves rpc reply from operation performed.

        Returns:
            str: rpc result from data_xml attribute.
        """
        return self.result.data_xml

    def _perform_operation(self) -> str:
        """
        Executes the specified operation on the device.

        Args: operation: The operation to perform.
        Returns: The result of the operation.
        """
        with self._establish_connection() as session:
            match self.operation:
                case "fetch":
                    return session.get(self.netconf_filter)

    def _establish_connection(self) -> manager:
        """
        Establishes a Netconf session with the device.

        Returns: The established session.
        """
        return manager.connect(
            host=self.device.host,
            port=self.device.port,
            username=self.device.username,
            password=self.device.password,
            hostkey_verify=self.device.hostkey_verify,
            device_params=self.device.device_params,
            timeout=self.device.timeout,
            allow_agent=self.device.allow_agent,  # use for old ssh servers, like XR 7.3.2
            look_for_keys=self.device.look_for_keys,  # use for old ssh servers, like XR 7.3.2
        )



from typing import Optional
from dataclasses import dataclass
from ncpeek.netconf_parsers import Parser
from ncpeek.netconf_devices import NetconfDevice


def _calculate_percentage(entry: dict) -> int:
    return (int(entry["used-memory"]) / int(entry["total-memory"])) * 100


@dataclass
class CiscoIOSXEMemoryParser(Parser):
    """
    A parser for memory data for Cisco IOSXE devices.
    """

    netconf_filter_id: Optional[str] = None
    device: Optional[NetconfDevice] = None

    def parse(
        self,
        data_to_parse: dict,
        device: NetconfDevice,
        netconf_filter_id: str,
    ) -> list[dict]:
        """
        Parse memory data.

        Args:
            data_to_parse (Dict): The data to parse.
            net_device (NetconfDevice): The device the data is related to.
            netconf_filter_id (str): The filter ID used for netconf.

        Returns:
            List[Dict]: The parsed data.
        """
        self.device = device
        self.netconf_filter_id = netconf_filter_id
        return self._extract_memory_statistics(rpc_reply=data_to_parse)

    def _extract_memory_statistics(self, rpc_reply: dict) -> str:
        stats: list = []
        xpath: dict = rpc_reply["data"]["memory-statistics"][
            "memory-statistic"
        ]

        for entry in xpath:
            stats.append(self._get_memory_stats(entry=entry))
        return stats

    def _get_memory_stats(self, entry: dict) -> dict:
        return {
            "name": entry["name"],
            "percent_used": _calculate_percentage(entry),
            "field": self.netconf_filter_id,
            "device": self.device.hostname,
            "ip": self.device.host,
        }



import sys
import argparse
from argparse import RawTextHelpFormatter

NCPEEK_DESCRIPTION = """'ncpeek' is a netconf client designed to fetch data from various devices.
The client can be utilized in two distinct ways, 
either via Command Line Interface (CLI) or Application Programming Interface (API).
The data retrieval can be filtered through either XML or XPath, 
however, only one filter type can be applied at a given time.
Note that in CLI mode, only filenames can be treated as arguments.
Source code: https://github.com/jillesca/ncpeek"""

DEVICE_SETTINGS_DESCRIPTION = """Specify JSON filename containing device settings.
Visit https://github.com/jillesca/ncpeek/tree/main/ncpeek/devices for more information."""

XML_FILTER_DESCRIPTION = """Specify XML filename containing XML filter.
Visit https://github.com/jillesca/ncpeek/tree/main/ncpeek/filters for more details."""

XPATH_FILTER_DESCRIPTION = """Formats: <xpath> OR <namespace>:<xpath>
Example: 'interfaces/interface' OR 
'http://cisco.com/ns/yang/Cisco-IOS-XE-interfaces-oper:interfaces/interface'"""


class CustomArgumentParser(argparse.ArgumentParser):
    """
    Custom ArgumentParser class to display help message on error.
    """

    def error(self, message: str) -> None:
        sys.stderr.write(f"Error: {message}")
        self.print_help()
        sys.exit(2)


def create_argument_parser():
    """
    Create and return an argument parser with mutually exclusive arguments.
    For device settings and filter options.
    """
    parser = CustomArgumentParser(
        description=NCPEEK_DESCRIPTION, formatter_class=RawTextHelpFormatter
    )
    group = parser.add_mutually_exclusive_group(required=True)

    parser.add_argument(
        "-d",
        "--device-settings",
        help=DEVICE_SETTINGS_DESCRIPTION,
    )

    group.add_argument(
        "-x",
        "--xml-filter",
        help=XML_FILTER_DESCRIPTION,
    )

    group.add_argument(
        "-p",
        "--xpath-filter",
        help=XPATH_FILTER_DESCRIPTION,
    )

    return parser



import re


def extract_xpath(netconf_filter: str) -> tuple:
    """
    Extract xpath from a netconf filter string.

    Args: netconf_filter (str): The netconf filter string.

    Returns: tuple: A tuple containing the type of filter and the extracted xpath.
    """
    if ":" not in netconf_filter:
        return ("xpath", netconf_filter)

    if "http://" in netconf_filter or "https://" in netconf_filter:
        return _extract_xpath_from_url(netconf_filter=netconf_filter)

    return _extract_xpath_without_url(netconf_filter=netconf_filter)


def _extract_xpath_from_url(netconf_filter: str) -> tuple:
    """
    Extract xpath from a netconf filter string that contains a URL.

    Args: netconf_filter (str): The netconf filter string.

    Returns: tuple: A tuple containing the type of filter and the extracted xpath.
    """
    ns_xpath = re.search(r"(https*:\/\/\S+):(\S+)", netconf_filter)
    ns = ns_xpath.group(1)
    select = ns_xpath.group(2)
    return ("xpath", ({"ns0": ns}, select))


def _extract_xpath_without_url(netconf_filter: str) -> tuple:
    """
    Extract xpath from a netconf filter string that does not contain a URL.

    Args: netconf_filter (str): The netconf filter string.

    Returns: tuple: A tuple containing the type of filter and the extracted xpath.
    """
    ns, select = netconf_filter.split(":")
    return ("xpath", ({"ns0": ns}, select))



from typing import List
from abc import ABC, abstractmethod
from ncpeek.netconf_devices import NetconfDevice


class Parser(ABC):
    """
    This is an abstract base class for parsers. It enforces the implementation
    of the 'parse' method in any derived class.
    """

    @classmethod
    @abstractmethod
    def parse(
        cls,
        data_to_parse: dict,
        net_device: NetconfDevice,
        netconf_filter_id: str,
    ) -> List:
        """
        This is an abstract method that must be implemented in any derived class.
        It should parse the given data and return a list.

        Args:
            data_to_parse: The data to parse.
            net_device: The network device.
            netconf_filter_id: The filter ID for netconf.

        Returns: A list of parsed data.
        """
        raise NotImplementedError(
            "Subclasses must implement this parse method."
        )



from typing import Optional, Union
from dataclasses import dataclass, field


@dataclass
class NetconfDevice:
    """
    This class represents a network device with its configurations.
    """

    host: str = None
    password: str = None
    username: str = None
    device_params: dict = field(default_factory=dict)
    hostname: Optional[str] = None
    port: Optional[int] = 830
    timeout: Optional[int] = 60
    hostkey_verify: Optional[Union[bool, str]] = False
    allow_agent: Optional[Union[bool, str]] = True
    look_for_keys: Optional[Union[bool, str]] = True

    def __post_init__(self):
        """
        Converts some attributes' string values to boolean.
        """
        self._check_required_fields()
        self.hostkey_verify = self.convert_str_to_bool(
            str_value=self.hostkey_verify
        )
        self.allow_agent = self.convert_str_to_bool(str_value=self.allow_agent)
        self.look_for_keys = self.convert_str_to_bool(
            str_value=self.look_for_keys
        )
        self._set_hostname()

    def _set_hostname(self) -> None:
        """
        Sets the hostname of the device if it is not already set.
        """
        if not self.hostname:
            self.hostname = self.host

    def _check_required_fields(self) -> None:
        """
        Validates that the required fields are provided.
        """
        missing_fields = []
        host = ""
        if not self.host:
            missing_fields.append("'Host'")
        else:
            host = self.host

        if not self.password:
            missing_fields.append("'Password'")
        if not self.username:
            missing_fields.append("'Username'")

        if missing_fields:
            fields = ", ".join(missing_fields)
            host_msg = f" for host '{host}'" if host else ""
            raise ValueError(
                f"Missing field(s){host_msg}: {fields}. These must be provided in JSON format."
            )

    @staticmethod
    def convert_str_to_bool(str_value: Union[bool, str]) -> bool:
        """
        This function parses boolean values from string representation.
        Needed as a string "False" in JSON is considered True in Python.
        """
        if isinstance(str_value, bool):
            return str_value
        if str_value.lower() == "true":
            return True
        if str_value.lower() == "false":
            return False
        else:
            raise ValueError(
                f"Invalid boolean value: {str_value} for NetconfDevice option."
            )
